import pandas as pd
from typing import Optional, List, Dict, Literal, Hashable, Any, Tuple
from pathlib import Path
import warnings
import copy
from pandas.api.types import is_string_dtype

import polars as pl

from tidydata.dtype import DtypeMappings, DtypeStr, PANDASDTYPE_TO_ALIAS
from tidydata.utils import (
    is_default_index,
    set_range,
    convert_dtype_backend,
    free_astype,
    free_fillna,
    log_error,
)


warnings.simplefilter(action="ignore", category=FutureWarning)
# pd.options.mode.use_inf_as_na = True
pd.options.mode.copy_on_write = True


class TableBase(object):
    """TableBase类"""

    def _repr_html_(self):
        return self.data._repr_html_()

    def __getitem__(self, key):
        return self.data[key]

    @classmethod
    def from_file(
        cls,
        path: Path | str,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        description: Optional[str] = None,
        columns: Optional[List[str]] = None,
        column_aliases: Optional[Dict[str, str]] = None,
        column_labels: Optional[Dict[str, str]] = None,
        column_dtypes: Optional[Dict[str, DtypeStr]] = None,
        value_repls: Optional[Dict[str, Dict[Hashable, Any]]] = None,
        value_ranges: Optional[Dict[str, Tuple | List | Dict[int, str]]] = None,
        value_nafills: Optional[Dict[str, Any]] = None,
        **options,
    ):
        from tidydata.io import read_file

        source = Path(path)

        obj = read_file(source, columns=columns, **options)

        if isinstance(obj, TableBase):
            data = obj.data
            name = name or obj.name
            tag = tag or obj.tag
            description = description or obj.description
        else:
            data = obj

        return cls(
            data=data,
            source=source,
            name=name,
            tag=tag,
            description=description,
            column_aliases=column_aliases,
            column_labels=column_labels,
            column_dtypes=column_dtypes,
            value_repls=value_repls,
            value_ranges=value_ranges,
            value_nafills=value_nafills,
        )

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | Dict[str, List | pd.Index | pd.Series],
        name: str,
        tag: str,
        description: str,
        source: Path | str,
        columns: Optional[List[str]] = None,
        column_labels: Optional[Dict[str, str]] = None,
        column_dtypes: Optional[Dict[str, DtypeStr]] = None,
        column_aliases: Optional[Dict[str, str]] = None,
        value_repls: Optional[Dict[str, Dict[Hashable, Any]]] = None,
        value_ranges: Optional[Dict[str, Tuple | List | Dict[int, str]]] = None,
        value_nafills: Optional[Dict[str, Any]] = None,
    ) -> None:
        if isinstance(data, pl.DataFrame):
            self.data = (
                data.to_pandas(use_pyarrow_extension_array=True)
                if columns is None
                else data.select(columns).to_pandas(use_pyarrow_extension_array=True)
            )

        else:
            self.data = pd.DataFrame(data, columns=columns)

            if not is_default_index(self.data.index):
                log_error(
                    f"Invalid Index: DataFrame in 'data' is not default range index, please use 'reset_index()' in pandas"
                )
            if isinstance(self.data.columns, pd.MultiIndex):
                log_error(
                    f"Invalid Column names: DataFrame column names in 'data' has multiple index, please flatten it"
                )

            if not is_string_dtype(self.data.columns):
                log_error(
                    f"Invalid Column names: DataFrame column names in 'data' is not string dtype"
                )

        self.name = name
        self.tag = tag
        self.description = description
        self.source = source
        self._column_labels = self.data.columns.to_series()

        (
            self.label_column_names(column_labels)
            .fill_column_missings(value_nafills)
            .replace_column_values(value_repls)
            .convert_column_dtypes(column_dtypes)
            .limit_column_ranges(value_ranges)
            .alias_column_names(column_aliases)
        )

    def copy(self):
        return copy.deepcopy(self)

    @property
    def index(self):
        return self.data.index

    @property
    def size(self):
        return self.data.size

    @property
    def columns(self):
        return self.data.columns

    @property
    def column_labels(self):
        return self._column_labels.reindex(self.columns).combine_first(
            self.column_names
        )

    @property
    def column_names(self):
        return self.data.columns.to_series()

    @property
    def column_dtypes(self):
        return self.data.dtypes.replace(PANDASDTYPE_TO_ALIAS)

    @property
    def column_ranges(self):
        result = {
            series.name: (
                (series.min(), series.max())
                if series.dtype.is_numeric() or series.dtype.is_temporal()
                else dict(enumerate(series.cat.get_categories()))
                if series.dtype == pl.Categorical
                else series.drop_nulls().unique().to_list()
                if series.dtype != pl.Null
                else []
            )
            for series in pl.from_pandas(self.data).iter_columns()
        }

        return pd.Series(result)

    @property
    def column_missings(self):
        return pd.Series(
            pl.from_pandas(self.data).null_count().transpose()["column_0"],
            index=self.data.columns,
        )

    def reset_table_name(self, value: Optional[str] = None):
        if value is not None:
            self.name = value
        return self

    def reset_table_tag(self, value: Optional[str] = None):
        if value is not None:
            self.tag = value
        return self

    def reset_table_description(self, value: Optional[str] = None):
        if value is not None:
            self.description = value
        return self

    def reset_table_source(self, value: Optional[str] = None):
        if value is not None:
            self.source = value

        return self

    def label_column_names(
        self,
        mapping: Optional[Dict[str, str]] = None,
    ):
        if mapping is not None:
            self._column_labels = (
                pd.Series(mapping)
                .reindex(self.data.columns)
                .combine_first(self._column_labels)
            )
        return self

    def replace_column_values(
        self,
        mapping: Optional[Dict[str, Dict[Any, Any]]] = None,
    ):
        if mapping is not None:
            self.data.replace(mapping, inplace=True)
        return self

    def fill_column_missings(
        self,
        mapping: Optional[Dict[str, Any]] = None,
    ):
        if mapping is not None:
            self.data = self.data.transform(
                lambda x: free_fillna(x, mapping.get(x.name)), axis=0
            )
        return self

    def convert_column_dtypes(
        self,
        mapping: Optional[Dict[str, DtypeStr]] = None,
    ):
        if mapping is not None:
            mapping = pd.Series(mapping).replace(DtypeMappings).to_dict()
            self.data = self.data.transform(
                lambda x: free_astype(x, mapping.get(x.name)), axis=0
            )
        return self

    def limit_column_ranges(
        self,
        mapping: Optional[Dict[str, Tuple | List | Dict[int, str]]] = None,
    ):
        if mapping is not None:
            self.data = self.data.apply(
                lambda x: set_range(x, mapping.get(x.name)), axis=0
            )
        return self

    def alias_column_names(
        self,
        mapping: Optional[Dict[str, str]] = None,
    ):
        if mapping is not None:
            self._column_labels.index = self._column_labels.index.map(
                lambda x: mapping.get(x, x)
            )
            self.data.rename(columns=mapping, inplace=True)
        return self

    def select_columns(
        self,
        columns: Optional[List[str] | str] = None,
    ):
        if columns is not None:
            if isinstance(columns, list):
                self.data = self.data[columns]
            else:
                self.data = self.data.filter(regex=columns)
            self._column_labels = self._column_labels.reindex(self.data.columns)
        return self

    def select_rows(
        self,
        rows: Optional[str | int | List[str]] = None,
    ):
        if rows is not None:
            if isinstance(rows, str):
                self.data = self.data.query(rows)
            elif isinstance(rows, list):
                self.data = self.data.query(
                    "&".join([f"({row.strip()})" for row in rows])
                )
            else:
                self.data = self.data.iloc[0:rows]

        return self

    def convert_dtype_backend(
        self,
        value: Optional[Literal["numpy", "numpy_nullable", "pyarrow"]] = None,
    ):
        if value is None:
            return self

        self.data = convert_dtype_backend(self.data)

        return self

    def drop_duplicate_values(
        self, columns: Optional[List[str] | Literal["all"]] = None
    ):
        if columns is None:
            return self

        columns = self.data.columns if columns == "all" else columns

        self.data.drop_duplicates(subset=columns, ignore_index=True, inplace=True)

        return self



import pandas as pd
from pandas.api.types import (
    is_any_real_numeric_dtype,
    is_integer_dtype,
    is_string_dtype,
    is_bool_dtype,
    is_numeric_dtype,
    is_datetime64_any_dtype,
    is_float_dtype
)
from typeguard import typechecked
from typing import Optional, Literal, Any, List
import re
from pandas._libs.missing import NAType
from numpy import log, log10, log2


@pd.api.extensions.register_series_accessor("bools")
class BooleanAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        if not is_bool_dtype(obj) or obj.dtype.name != "bool[pyarrow]":
            raise TypeError(f"Not a boolean series with dtype name 'bool[pyarrow]'.")

    def to_dummy(self, fill_value: Optional[bool] = None):
        bool_ss = self._obj
        if fill_value is not None:
            bool_ss = bool_ss.fillna(fill_value)

        return (
            bool_ss.astype(object).replace({True: 1, False: 0}).astype("int64[pyarrow]")
        )

    def to_cats(
        self,
        true: str = "Yes",
        false: str = "No",
        fill_value: Optional[bool] = None,
        ordered: bool = True,
    ):
        bool_ss = self._obj
        if fill_value is not None:
            bool_ss = bool_ss.fillna(fill_value)

        return (
            bool_ss.astype(object)
            .replace({True: true, False: false})
            .astype(pd.CategoricalDtype([false, true], ordered=ordered))
        )

    def _or(self, other: pd.Series, skipna: bool = True):
        if not is_bool_dtype(other):
            raise TypeError(f"Not a boolean series in 'other': {other}.")

        xor_ss = (self._obj | other).fillna(False)
        if skipna:
            both_nas = self._obj.isna() & other.isna()
            return xor_ss.mask(both_nas, pd.NA)
        else:
            any_nas = self._obj.isna() | other.isna()
            return xor_ss.mask(any_nas, pd.NA)


@pd.api.extensions.register_series_accessor("cats")
class CategoricalAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        if not isinstance(obj.dtype, pd.CategoricalDtype):
            raise TypeError(f"Not a Categorical series.")

    def to_dummy(
        self, true: List[str] | str | pd.Series, fill_value: Optional[Any] = None
    ):
        cat_ss = self._obj
        cat_codes = cat_ss.cat.categories

        if fill_value is not None:
            if fill_value not in cat_codes:
                raise ValueError(
                    f"Value in 'fill_value' is not in existing categories."
                )
            else:
                cat_ss = cat_ss.fillna(fill_value)

        if isinstance(true, pd.Series) and not is_bool_dtype(true):
            raise TypeError(f"Value in 'true' is not a boolean series: {true}.")

        if isinstance(true, pd.Series):
            fill_true = true.fillna(False)
            return (
                self._obj.astype(object)
                .where(fill_true, 0)
                .mask(fill_true, 1)
                .where(true.notna(), pd.NA)
                .astype("int64[pyarrow]")
            )
        else:
            if isinstance(true, str):
                true = [true]
            repls = {cat: 1 if cat in set(true) else 0 for cat in cat_codes}
            return cat_ss.astype(object).replace(repls).astype("int64[pyarrow]")

    def to_bool(
        self, true: List[str] | str | pd.Series, fill_value: Optional[Any] = None
    ):
        return self._obj.cats.to_dummy(true, fill_value).astype("bool[pyarrow]")



@pd.api.extensions.register_series_accessor("combs")
class CombineAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def _max(self, other: pd.Series, skipna: bool = True):
        return pd.concat((self._obj, other), axis=1).max(axis=1, skipna=skipna)

    def _min(self, other: pd.Series, skipna: bool = True):
        return pd.concat((self._obj, other), axis=1).min(axis=1, skipna=skipna)

    def _mean(self, other: pd.Series, skipna: bool = True):
        return pd.concat((self._obj, other), axis=1).mean(axis=1, skipna=skipna)
    
    def _if(self, other: pd.Series, cond: pd.Series, skipna: bool = False):
        fill_cond = cond.fillna(False)
        mask_cond = (
            fill_cond | self._obj.isna() if skipna else fill_cond & self._obj.notna()
        )
        return self._obj.mask(mask_cond, other)


@pd.api.extensions.register_series_accessor("strs")
class StringAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        if obj.dtype.name != "string":
            raise TypeError(f"Not a string series with dtype name 'string'.")

    def to_numeric(self, coerce: bool = False):
        errors = "coerce" if coerce else "raise"
        return pd.to_numeric(self._obj, errors=errors, dtype_backend="pyarrow")

    def cat(
        self, other: str | pd.Series, sep: str = "", fill_value: Optional[str] = None
    ):
        ss = self._obj.fillna(fill_value) if fill_value is not None else self._obj

        if isinstance(other, str):
            return ss + sep + other
        else:
            return ss.str.cat(other, sep=sep, na_rep=fill_value)

    def lit(
        self, other: str
    ):
        return pd.Series(other,index=self._obj.index,dtype='string[pyarrow]')

@pd.api.extensions.register_series_accessor("ints")
class IntegerAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        if not is_integer_dtype(obj):
            raise TypeError(f"Not a integer series: {obj}.")

    def to_string(self):
        return self._obj.astype("string[pyarrow]")

    def log(self, base: Literal["e"] = "e"):
        log_funcs = {"e": log, "10": log10, "2": log2}
        return (
            self._obj.astype("float64")
            .transform(log_funcs[base])
            .astype("float64[pyarrow]")
        )



@pd.api.extensions.register_series_accessor("floats")
class IntegerAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _validate(self, obj):
        if not is_float_dtype(obj):
            raise TypeError(f"Not a float series: {obj}.")
        
    def log(self, base: Literal["e"] = "e"):
        log_funcs = {"e": log, "10": log10, "2": log2}
        return (
            self._obj.astype("float64")
            .transform(log_funcs[base])
            .astype("float64[pyarrow]")
        )

@pd.api.extensions.register_series_accessor("econ")
class EconomicsAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self._educ10_cn = {
            "No formal education illiterate": 1,
            "Did not finish primary school but capable of reading and/or writing": 2,
            "Sishu": 3,
            "Elementary school": 6,
            "Middle school": 9,
            "High school": 12,
            "Vocational school": 12,
            "Two/Three Year College/Associate degree": 15,
            "Four Year College/Bachelor's degree": 16,
            "Post-graduated(Master/PhD)": 19,
        }
        

        
        self._educ_system = {"cn10": self._educ10_cn}

    def to_educ_year(self, system: Optional[Literal["cn10"]] = "cn10"):
        ss = self._obj
        if not isinstance(
            ss.dtype, pd.CategoricalDtype
        ) and not ss.dtype.name.startswith("dictionary"):
            raise TypeError(f"Not a categorical or dictionary series: {ss}.")

        if ss.dtype.name.startswith("dictionary"):
            ss = ss.astype("category")

        if set(ss.cat.categories) != set(self._educ_system[system].keys()):
            raise ValueError(
                f"Categories Mismatch: categories from series can not match the categoeies of 'cn10'."
            )

        return (
            ss.astype(object)
            .replace(self._educ_system[system])
            .astype("int64[pyarrow]")
        )
    def to_educ_level(self, system: Optional[Literal['cn6','isced97-3','cn8','cn2']]="cn6"):
        
        cn10_to_cn8 = {
            "No formal education illiterate": 'Illiterate',
            "Did not finish primary school but capable of reading and/or writing": 'Capable of reading and/or writing',
            "Sishu": 'Capable of reading and/or writing',
            "Elementary school": 'Elementary school',
            "Middle school": "Middle school",
            "High school": "High school/Vocational school",
            "Vocational school": "High school/Vocational school",
            "Two/Three Year College/Associate degree": "Two/Three Year College/Associate degree",
            "Four Year College/Bachelor's degree": "Four Year College/Bachelor's degree",
            "Post-graduated(Master/PhD)":"Post-graduated(Master/PhD)"
        }
        cn10_to_cn6 = {
            "No formal education illiterate": 'Less than middle school',
            "Did not finish primary school but capable of reading and/or writing": 'Less than middle school',
            "Sishu": 'Less than middle school',
            "Elementary school": 'Less than middle school',
            "Middle school": "Middle school",
            "High school": "High school/Vocational school",
            "Vocational school": "High school/Vocational school",
            "Two/Three Year College/Associate degree": "Two/Three Year College/Associate degree",
            "Four Year College/Bachelor's degree": "Four Year College/Bachelor's degree",
            "Post-graduated(Master/PhD)":"Post-graduated(Master/PhD)"
        }
        
        cn10_to_isced3 = {
            "No formal education illiterate": 'Less than higher secondary education',
            "Did not finish primary school but capable of reading and/or writing":'Less than higher secondary education',
            "Sishu": 'Less than higher secondary education',
            "Elementary school": 'Less than higher secondary education',
            "Middle school": 'Less than higher secondary education',
            "High school": "Higher secondary education",
            "Vocational school": "Higher secondary education",
            "Two/Three Year College/Associate degree": "Tertiary eduction",
            "Four Year College/Bachelor's degree": "Tertiary eduction",
            "Post-graduated(Master/PhD)":"Tertiary eduction"
        }
        
        cn10_to_cn2 = {
            "No formal education illiterate": 'Without tertiary education',
            "Did not finish primary school but capable of reading and/or writing":'Without tertiary education',
            "Sishu": 'Without tertiary education',
            "Elementary school": 'Without tertiary education',
            "Middle school": 'Without tertiary education',
            "High school": 'Without tertiary education',
            "Vocational school": 'Without tertiary education',
            "Two/Three Year College/Associate degree": "Tertiary eduction",
            "Four Year College/Bachelor's degree": "Tertiary eduction",
            "Post-graduated(Master/PhD)":"Tertiary eduction"
        }
        
        if system == 'isced97-3':
            cat_dtype = pd.CategoricalDtype(
                ['Less than higher secondary education',
                'Higher secondary education',
                'Tertiary eduction'],ordered=True)
            
            return self._obj.astype('string[pyarrow]').replace(cn10_to_isced3).astype(cat_dtype)

        elif system == 'cn2':
            cat_dtype = pd.CategoricalDtype(
                ['Without tertiary education',
                'Tertiary eduction'],ordered=True)
            
            return self._obj.astype('string[pyarrow]').replace(cn10_to_cn2).astype(cat_dtype)


        elif system == 'cn8':
            cat_dtype = pd.CategoricalDtype(
                ['Illiterate',
                'Capable of reading and/or writing',
                'Elementary school',
                'Middle school',
                'High school/Vocational school',
                'Two/Three Year College/Associate degree',
                "Four Year College/Bachelor's degree",
                "Post-graduated(Master/PhD)"],ordered=True)

            return self._obj.astype('string[pyarrow]').replace(cn10_to_cn8).astype(cat_dtype)
        
        else:
            cat_dtype = pd.CategoricalDtype(
                ['Less than middle school',
                'Middle school',
                'High school/Vocational school',
                'Two/Three Year College/Associate degree',
                "Four Year College/Bachelor's degree",
                "Post-graduated(Master/PhD)"],ordered=True)

            return self._obj.astype('string[pyarrow]').replace(cn10_to_cn6).astype(cat_dtype)
    
    def prov_to_area(self, system: Literal['econ3']='econ3'):
        """转换省份名到大区域名"""
        
        # 国家统计局2011经济分区 
        # http://www.stats.gov.cn/ztjc/zthd/sjtjr/dejtjkfr/tjkp/201106/t20110613_71947.htm
        prov_to_econ3 = {
            "北京":"东部",
            "天津":"东部",
            "河北":"东部",
            "上海":"东部",
            "江苏":"东部",
            "浙江":"东部",
            "福建":"东部",
            "山东":"东部",
            "广东":"东部",
            "海南":"东部",
            "辽宁":"东部",
            "吉林":"东部",
            "黑龙江":"东部",
            "山西":"中部",
            "安徽":"中部",
            "江西":"中部",
            "河南":"中部",
            "湖北":"中部",
            "湖南":"中部",
            "内蒙古":"西部",
            "广西":"西部",
            "重庆":"西部",
            "四川":"西部",
            "贵州":"西部",
            "云南":"西部",
            "西藏":"西部",
            "陕西":"西部",
            "甘肃":"西部",
            "青海":"西部",
            "宁夏":"西部",
            "新疆":"西部"
        }
        
        if system == "econ3":
            cat_dtype = pd.CategoricalDtype(['东部','中部','西部'],ordered=True)
            return self._obj.astype('string[pyarrow]').replace(prov_to_econ3).astype(cat_dtype)
        



import pandas as pd
from typing import (
    Optional,
    Tuple,
    List,
    Dict,
    Literal,
    Iterable,
    Any,
    Mapping,
)
from typeguard import typechecked
import platform
from pandas.api.types import is_string_dtype, is_object_dtype, infer_dtype
import polars as pl
import re
from collections import Counter
from loguru import logger


def strflatten_index(idx: pd.Index, sep: str = "_", fillna: str = "NA") -> pd.Index:
    """Normalize to a string, flat, and no-NA (or No space) index"""

    if isinstance(idx, pd.MultiIndex):
        return idx.to_flat_index().fillna(fillna).map(lambda x: sep.join(map(str, x)))
    else:
        return idx.fillna(fillna).astype(str)


def set_range(series: pd.Series, range: Optional[Tuple | List | Dict[int, str]] = None):
    if range is None:
        return series

    if isinstance(range, tuple) and len(range) > 2 or len(range) == 0:
        raise ValueError(
            f"Invalid value in mapping item: the length of tuple range {range} are greater than 2"
        )

    if isinstance(range, tuple):
        if len(range) == 1 and isinstance(range[0], tuple):
            return series.clip(range[0], range[1], inplace=True)
        else:
            return series.where(series.between(range[0], range[1], inclusive="both"))

    if isinstance(range, list):
        return series.where(series.isin(range))

    if isinstance(range, dict):
        if isinstance(series.dtype, pd.CategoricalDtype):
            categories = list(dict(sorted(range.items())).values())

            return series.astype(pd.CategoricalDtype(categories, ordered=True))
        else:
            raise TypeError(
                f"Unsupported Dtype: dict range '{range}' can not apply to non-categorical series"
            )


def free_fillna(series: pd.Series, value: Any):
    if value is None:
        return series

    try:
        return series.fillna(value)
    except:
        return series.astype(object).fillna(value)


def free_astype(series: pd.Series, dtype: Any):
    if dtype is None:
        return series
    try:
        return series.astype(dtype)
    except:
        return series.astype(object).astype(dtype)


def is_default_index(idx: pd.Index) -> bool:
    """Check whether is default index"""
    return isinstance(idx, pd.RangeIndex) and idx.start == 0 and idx.step == 1


def infer_stata_bin(version: int = 18):
    system = platform.system()
    if system == "Linux":
        return rf"/usr/local/stata{version}"
    elif system == "Windows":
        return rf"C:/Program Files/Stata{version}"
    elif system == "Darwin":
        return rf"/Applications/Stata"
    else:
        raise Exception(f"Unsupported System: {system}")


def cast_mix_dtype(
    df: pd.DataFrame, dtype_backend: Literal["numpy_nullable", "pyarrow"] = "pyarrow"
):
    to_string_type = "string[pyarrow]" if dtype_backend == "pyarrow" else "string"

    return df.transform(
        lambda x: x.astype(to_string_type)
        if infer_dtype(x).startswith("mix")
        and is_object_dtype(x)
        and not is_string_dtype(x)
        or (
            isinstance(x.dtype, pd.CategoricalDtype)
            and not is_string_dtype(x.cat.categories)
        )
        else x,
        axis=0,
    ).convert_dtypes(dtype_backend=dtype_backend)


def convert_dtype_backend(
    df: pd.DataFrame, backend: Literal["numpy", "numpy_nullable", "pyarrow"] = "pyarrow"
):
    if backend == "numpy":
        return pl.from_pandas(df).to_pandas()
    else:
        return df.convert_dtypes(dtype_backend=backend)


def lreshape_frame(
    df: pd.DataFrame,
    groups: List[str] | Dict[str, List[str]],
    i: Optional[List[str] | str] = None,
    j: Optional[str] = None,
    sep: str = "@",
    suffix: str = r".*",
    dropna: bool = False
) -> pd.DataFrame:
    """Reshape to long format"""

    if not df.columns.is_unique:
        raise IndexError(f"column names is not unique")

    if j is None:
        long_df = pd.lreshape(df, groups, dropna=True)
    else:
        widecol_pattern = re.compile(f"({'|'.join(set(groups))})({sep})({suffix})")
        i = (
            df.columns.where(~df.columns.str.fullmatch(widecol_pattern)).dropna()
            if i is None
            else i
        )
        long_df = (
            pd.wide_to_long(df, stubnames=groups, i=i, j=j, sep=sep, suffix=suffix)
            .reset_index()
            .astype({j: "string[pyarrow]"})
        )

        if long_df[i].empty:
            raise ValueError(
                f"Can not match wide columns with pattern sep and suffix: {sep}{suffix}"
            )
    if dropna:
        subset = groups.keys() if isinstance(groups,dict) else groups
        return long_df.dropna(how='all', subset=subset)
    else:
        return long_df 


def wreshape_frame(
    df: pd.DataFrame,
    groups: Optional[List[str] | Dict[str, List[str]]] = None,
    i: Optional[List[str] | str] = None,
    j: Optional[str] = None,
    sep: str = "@",
) -> pd.DataFrame:
    """Reshape to wide format"""

    if not isinstance(groups, dict) and j is None:
        raise ValueError(f"'groups' is not a dictionary while 'j' is None")
    if isinstance(groups, dict) and j is not None:
        raise ValueError(f"'groups' and 'j' is conflict, please set one of them")

    columns = list(groups.keys()) if isinstance(groups, dict) else j
    values = list(groups.values()) if isinstance(groups, dict) else groups

    index = i or list(
        set(df.columns)
        - set(values)
        - (
            set(columns)
            if isinstance(columns, list)
            else {columns}
            if isinstance(columns, str)
            else set()
        )
    )

    wide_df = df.pivot(columns=columns, index=index, values=values)

    wide_df.columns = strflatten_index(
        wide_df.columns,
        sep=sep,
        fillna="NA",
    )
    wide_df.reset_index(inplace=True)
    return wide_df


def reshape_frame(
    df: pd.DataFrame,
    to_shape: Literal["long", "wide"],
    groups: Optional[List[str] | Dict[str, List[str]]] = None,
    i: Optional[List[str] | str] = None,
    j: Optional[str] = None,
    sep: str = "@",
    suffix: str = r".*",
    dropna: bool = False
):
    if to_shape == "long":
        return lreshape_frame(df, groups, i, j, sep, suffix, dropna)
    else:
        return wreshape_frame(df, groups, i, j, sep)


def append_frames(
    dfs: Iterable[pd.DataFrame],
    on: Optional[List[str]] = None,
    how: Literal["inner", "outer"] = "outer",
    keycol: Optional[Dict[str, List[str] | List[int]]] = None,
) -> pd.DataFrame:
    """Append dataframes to a dataframe"""

    if on is not None:
        dfs = (df.filter(items=on, axis=1) for df in dfs)

    if keycol is not None:
        key_idx = pd.MultiIndex.from_frame(pd.DataFrame(keycol))
        names = key_idx.names
        concated_df = pd.concat(
            dfs, join=how, copy=False, keys=key_idx.to_list(), names=names
        )
        concated_df.index = concated_df.index.droplevel(
            [name for name in concated_df.index.names if name not in set(names)]
        )
        concated_df.reset_index(inplace=True)

        return concated_df.astype(
            {
                k: "string[pyarrow]" if isinstance(v[0], str) else "int64[pyarrow]"
                for k, v in keycol.items()
            }
        )
    else:
        return pd.concat(dfs, join=how, copy=False, ignore_index=True)


def join_frames(
    dfs: Iterable[pd.DataFrame],
    on: Optional[List[str]] = None,
    how: Literal["left", "inner", "outer"] = "outer"
):
    """Merge dataframes to a dataframe"""

    if on is not None:
        dfs = (df.set_index(on) for df in dfs)
        
    return next(dfs).join(dfs, how=how).reset_index()


def concat_frames(
    dfs: Iterable[pd.DataFrame],
    axis: Literal["column", "row"],
    on: Optional[List[str]] = None,
    how: Literal["left", "inner", "outer"] = "outer",
    keycol: Optional[Dict[str, List[str | int]]] = None,
):
    if axis == "column":
        return join_frames(dfs, on, how)
    else:
        return append_frames(dfs, on, how, keycol)


def merge_dict_in_nest_list(
    list_of_lists: List[List[Dict] | List[str]], allow_duplciate_key: bool = False
) -> List[Dict | str]:
    merged_list = []
    for sublist in list_of_lists:
        merged_dict = {}
        for d in sublist:
            for key in d:
                if not allow_duplciate_key and key in merged_dict:
                    raise ValueError(f"Duplicate key '{key}' found. Merging stopped.")
                merged_dict[key] = d[key]
        merged_list.append(merged_dict)
    return merged_list


def get_duplicate_items(items: Iterable | Mapping):
    counter = Counter(items)
    return [item for item, count in counter.items() if count > 1]


def log_error(message: str, *args, throw: bool = True, **kwargs):
    logger.error(message, *args, **kwargs)
    if throw:
        raise Exception(message)


@typechecked
def split_frame_by_columns(df: pd.DataFrame, by: str | List[str]):
    for k, v in df.groupby(by=by):
        if isinstance(k, tuple):
            yield f"{'_'.join(map(lambda x: 'NA' if x is pd.NA else str(x),k))}", v
        else:
            yield f"{k}", v


@typechecked
def split_frame_by_rows(df: pd.DataFrame, by: int):
    for start in range(0, df.index.size, by):
        end = start + by
        yield f"{start}", df[start:end]


def split_frame(
    df: pd.DataFrame,
    by_columns: Optional[List[str] | str] = None,
    by_rows: Optional[int] = None,
):
    if (by_columns or by_rows) is None:
        raise ValueError(f"'by_columns' and 'by_rows' should not be all None.")

    if by_columns is not None and by_rows is None:
        for k, v in split_frame_by_columns(df, by=by_columns):
            yield k, v

    elif by_columns is None and by_rows is not None:
        for k, v in split_frame_by_rows(df, by=by_rows):
            yield k, v

    else:
        for k, sub_df in split_frame_by_columns(df, by=by_columns):
            for start in range(0, sub_df.index.size, by_rows):
                end = start + by_rows
                yield f"{k}_{start}", sub_df[start:end]


# def gfill(ss: pd.Series, groupby: pd.Series | List[pd.Series], method: Literal['b','f','bf','fb'] = 'fb'):
    
#     df = pd.concat([ss] + groupby) if isinstance(groupby, list) else pd.concat([ss,groupby])
    
#     if method == 'b':
#         return df.groupby([])
#     pass


from typing import Literal

DtypeStr = Literal[
    "Int",
    "UInt",
    "Float",
    "Bool",
    "Str",
    "Cat",
    "Date",
    "Time",
    "Duration",
    "Obj",
    "Null",
    "Bin",
]

DtypeMappings = {
    "Int": "int64[pyarrow]",
    "UInt": "uint64[pyarrow]",
    "Float": "float64[pyarrow]",
    "Bool": "bool[pyarrow]",
    "Str": "string[pyarrow]",
    "Cat": "category",
    "Datetime": "datetime64[ns]",
    "Duration": "timedelta64[ns]",
    "Obj": "object",
    "Null": "null[pyarrow]",
    "Bin": "binary[pyarrow]",
}

PANDASDTYPE_TO_ALIAS = {
    "int64[pyarrow]": "Int",
    "int64": "Int",
    "Int64": "Int",
    "uint64[pyarrow]": "UInt",
    "float64[pyarrow]": "Float",
    "double[pyarrow]": "Float",
    "float64": "Float",
    "Float64": "Float",
    "bool[pyarrow]": "Bool",
    "string[pyarrow]": "Str",
    "category": "Cat",
    "datetime64[ns]": "Datetime",
    "timedelta64[ns]": "Duration",
    "object": "Obj",
    "null[pyarrow]": "Null",
    "binary[pyarrow]": "Bin",
}
